use std::{cell::RefCell, collections::HashMap};

use aes_gcm::aes::cipher::BlockEncrypt;
pub use ed25519_dalek::{Keypair, PublicKey, Signature, Signer};

use crate::prelude::*;

pub enum TrustLevel {
    Unknown,
    Distrusted,
    // Future extension
    // Transitively,
    Ultimately,
}

pub struct TrustDB {}

#[derive(serde::Serialize, serde::Deserialize, Default, Debug)]
pub enum PacketHeader {
    #[default]
    NoCrypto,
    Signed(Signature),
    // u64 here is the nonce
    Encrypted(u64),
}

pub struct SymmetricKey {
    key: [u8; 16],
    nonce: [u8; 16],
}

#[derive(Default)]
pub(crate) struct VerifyPacketSignature {}

impl Service for std::rc::Rc<VerifyPacketSignature> {
    fn setup(&self, api: &crate::Fleck) {
        api.channel::<fleck_core::ReceivePacketChannel>().sub_opt(
            fleck_core::ReceiveOrder::Verify,
            self,
            VerifyPacketSignature::check,
        );
    }
}

impl VerifyPacketSignature {
    fn check(
        &self,
        api: &crate::Fleck,
        mut message: fleck_core::msg::Message,
    ) -> Option<fleck_core::msg::Message> {
        match message.crypto_header {
            PacketHeader::Signed(signature) => {
                message.crypto_header = PacketHeader::NoCrypto;
                let to_verify = bincode::serialize(&message)
                    .expect("assume that we can re-serialize the message");

                // signature.verify(to_verify, signature);

                // let pubkey = api.services().with_service(|ns: &fleck_core::Nodes| { ns.self_node().pubkey() });

                // match pubkey {
                // msg.crypto_header = PacketHeader::Signed(keypair.sign(&sign_data));
                // }
            }
            _ => {}
        }

        Some(message)
    }
}

#[derive(Default)]
pub(crate) struct SignPacket {}

impl Service for std::rc::Rc<SignPacket> {
    fn setup(&self, api: &crate::Fleck) {
        api.channel::<fleck_core::SendPacketChannel>().sub_ref(
            fleck_core::SendOrder::Sign,
            self,
            SignPacket::sign,
        );
    }
}

impl SignPacket {
    fn sign(&self, api: &crate::Fleck, message: &mut fleck_core::msg::Message) {
        let params = message
            .saved_params
            .expect("packet got to signing stage without saved parameters?");

        // don't do anything if the packet doesn't want a signature
        if !params.signed {
            return;
        }

        api.with_service(|ns: &fleck_core::NodeService| {
            let keypair = ns.self_node().keypair().unwrap();
            let sign_data = bincode::serialize(&message.content).unwrap();
            message.crypto_header = PacketHeader::Signed(keypair.sign(&sign_data));
        });
    }
}

#[derive(serde::Serialize, serde::Deserialize)]
pub struct KeyExchange {
    key: x25519_dalek::PublicKey,
}

impl fleck_core::msg::MessageParams for KeyExchange {
    const NAME: &'static str = "KeyExchange";
    const ENCRYPTED: bool = false;
    const SIGNED: bool = true;
}

pub enum KeyExchangeState {
    NoState,
    Given(x25519_dalek::PublicKey),
    WaitForResponse(x25519_dalek::EphemeralSecret),
    Completed(x25519_dalek::PublicKey, SymmetricKey),
}

pub struct EncryptionService {
    state: RefCell<HashMap<fleck_core::peer::Peer, KeyExchangeState>>,
}

impl Default for EncryptionService {
    fn default() -> Self {
        Self {
            state: Default::default(),
        }
    }
}

impl Service for std::rc::Rc<EncryptionService> {
    fn setup(&self, api: &crate::Fleck) {
        api.channel::<fleck_core::node::NodeRegistrationChannel>().sub_ref(self, EncryptionService::new_node);
        api.channel::<fleck_core::msg::MessageChannel<KeyExchange>>().sub_eat(self, EncryptionService::incoming);
    }
}

impl EncryptionService {
    fn new_node(&self, api: &crate::Fleck, node: &mut std::rc::Rc<fleck_core::node::Node>) {
        match node.peer().as_ref() {
            Some(peer) => self.begin_handshake(api, peer),
            None => (),
        }
    }

    fn begin_handshake(&self, api: &crate::Fleck, with: &fleck_core::peer::Peer) {
        use rand::SeedableRng;
        let mut csprng = rand::rngs::StdRng::from_entropy();
        let secret = x25519_dalek::EphemeralSecret::new(&mut csprng);

        // send notification with new secret
        api.queue::<fleck_core::SendPacketChannel>(fleck_core::msg::Message::build(
            KeyExchange {
                key: x25519_dalek::PublicKey::from(&secret)
            }
        ));

        let mut states = self.state.borrow_mut();
        let state = states.remove(&with).unwrap_or(KeyExchangeState::NoState);
        match state {
            KeyExchangeState::NoState => {
                states.insert(with.clone(), KeyExchangeState::WaitForResponse(secret));
            },
            KeyExchangeState::Given(pubkey) => {
                let result = secret.diffie_hellman(&pubkey);

                let sk = SymmetricKey {
                    key: result.as_bytes()[0..16].try_into().unwrap(),
                    nonce: result.as_bytes()[16..32].try_into().unwrap(),
                };

                states.insert(with.clone(), KeyExchangeState::Completed(pubkey, sk));
            },
            KeyExchangeState::WaitForResponse(secret) => {
                states.insert(with.clone(), KeyExchangeState::WaitForResponse(secret));
            },
            KeyExchangeState::Completed(_pubkey, _symmetric_key) => {
                log::error!("asked to begin handshake when already have a good encryption key?");
            },
        }
    }

    fn incoming(&self, api: &crate::Fleck, msg: (fleck_core::msg::Metadata, KeyExchange)) {
        let mut states = self.state.borrow_mut();
        let peer = msg.0.peer.expect("incoming KeyExchange should have a peer?");
        // let state = states.entry(peer.clone()).or_insert(KeyExchangeState::NoState);
        let state = states.remove(&peer).unwrap_or(KeyExchangeState::NoState);
        match state {
            KeyExchangeState::NoState => {
                states.insert(peer.clone(), KeyExchangeState::Given(msg.1.key));

                drop(states);
                self.begin_handshake(api, &peer);
            },
            KeyExchangeState::Given(peer_pubkey) => {
                if peer_pubkey != msg.1.key {
                    log::info!("peer DH key changed!");
                    states.insert(peer.clone(), KeyExchangeState::Given(msg.1.key));
                }

                drop(states);
                self.begin_handshake(api, &peer);
            },
            KeyExchangeState::WaitForResponse(secret) => {
                // key handshake is finished!
                let result = secret.diffie_hellman(&msg.1.key);

                let sk = SymmetricKey {
                    key: result.as_bytes()[0..16].try_into().unwrap(),
                    nonce: result.as_bytes()[16..32].try_into().unwrap(),
                };

                states.insert(peer.clone(), KeyExchangeState::Completed(msg.1.key, sk));
            },
            KeyExchangeState::Completed(peer_pubkey, _symmetric_key) => {
                if peer_pubkey == msg.1.key {
                    // it's a repeat, so nothing to do
                    return;
                }
                // it's a new pubkey, so we want to restart the handshake
                states.insert(peer.clone(), KeyExchangeState::Given(msg.1.key));
                drop(states);
                self.begin_handshake(api, &peer);
            },
        }
    }
}
